      program pgom

! This is the main routine for a planetary geostrophic ocean
!  circulation model in a closed domain.

!  Authors: R. M. Samelson and G. K. Vallis, November 1995.
!    Revised: May 1996.

!  A description of the basic equations and the method of
!  numerical implementation can be found in
!   (i) Samelson, R. and Vallis, G. K. 1995: A simple friction and
!        diffusion scheme for planetary geostrophic basin models.
!         J. Phys. Oceanogr. (submitted)
!   (ii) Samelson, R. and Vallis, G. K. 1995: Large-scale circulation
!        with small diapycnal diffusion: the two-thermocline limit.
!         J. Mar. Res. (submitted).
!  This code implements 'Form A' of the planetary geostrophic (PG)
!   or `thermocline' equations from paper (i).  See (i) for a
!   theoretical comparison of these equations to other PG sets.
!  See the appendix of (ii) for a partial description of the finite
!   difference scheme.
!  Temperature is the default prognostic variable.  There is a
!   run-time switch to include salinity.  There is a stretched vertical
!   grid, and a run-time switch for Cartesian/spherical horizontal
!   coordinates.  Data I/O is normally implemented using NetCDF format,
!   although binary or even ascii I/O is possible.  We strongly
!   encourage the use of the default NetCDF I/O format whenever
!   possible, in order to allow sharing of output files and analysis
!   code.  A simple Makefile is include for UNIX users.

!    ***This code is not meant to be used as a black box***
!  The code was primarily developed over a period from Sept 1993
!   to Sept.  1995.  This version was made public in November 1995.
!   It is meant to be easy to understand, use, and modify, and the
!   user is expected to be familiar with it.  It is not fully
!   documented, and may not be completely robust, although it is
!   believed to be free of major bugs.  It is coded in standard
!   FORTRAN 77, except for: (i) the 'namelist' extension, (ii) use of
!   the do - enddo loop construction, and (iii) use of '!' to denote
!   comments.  If your compiler doesn't support these, we suggest you
!   get a new one!  It is coded in 'single precision' and with generic
!   function calls.  Use in real*8 is recommended, however, and this
!   can generally be achieved with compilation options.  It has been
!   run on various workstations (Dec Alpha, SGI, Sun etc.) and on a Cray
!   Y-MP, but not exhaustively tested.  A Fortran 90 version should
!   be available soon.

! The complete code consists of:
!  Fortran files:	pgmain.f, pginit.f, pguvwp.f, pgio_cdf.f,
!                       pgio_bin, pgother.f
!  Include files:	common.inc, params.inc
!  Namelist:		const.name

! pgmain.f:    Many of the main routines
! pginit.f:    Set-up and initialization routines 
! pguvwp.f:    Routines to calculate velocity from density and pressure
!               (including elliptic solver for barotropic pressure)
! pgio_cdf.f:  Data I/O routines (NetCDF format). 
! pgio_bin.f:  Data I/O routines (binary format). 
! pgother.f:   Additional routines, e.g., input with interpolation
!               to double resolution
! <echeckb.f:   Optional checking routines (usually not called and >
! <         currently out-of-date).     ***Not for distribution.***>

! As a general rule 'static' (unchanging) variables are held in 
! common blocks, whereas dynamic variables are passed via subroutine
! arguments. The common block is held in a file (common.inc)
! that is 'included' wherever it is needed. 

! Comments or questions should be directed to the authors
!   via email to rsamelson@whoi.edu or vallis@cascade.ucsc.edu
!
!  Grid dimensions:  NX,NY,NZ - east, north, up
!

!  Partial outline of calling sequence:
!       pgmain:  main calling program
!         initlz:  Set constants and initialize arrays
!          pgin:   Read initial data, if any
!          init1:  Set forcing and initial values
!          ell2d:  Solve elliptic problem for barotropic velocity
!         --begin main timestepping loop
!          pgstep:  Do one t-s step (Runge Kutta/Euler Forward)
!           pguvw:  Get velocities
!           boundb: Lateral boundary condition (no-normal-flow) solver
!           pgadv:  Density and salinity advection
!           ekman:  Ekman advection and surface flux
!           fluxbc: Bottom boundary condition
!           pgvdf:  Vertical diffusion
!           pghdf:  Horizontal diffusion
!           eustpr: Euler forward step
!           leapfr: leapfrog step
!         --end timestepping loop
!         pgout:  Write pg arrays
!

! Arrays contain the following variables
	! t, to		Temperature (at two timesteps)
	! s, so		Salinity  (at two timesteps)
	! r		Density
	! p0		Barotropic Pressure
	! p		Pressure
	! u, v, w	Velocity fields
	! tt (xrt)      Right hand side of T evolution equation
	! st (xrt)      Right hand side of S evolution equation

! Let's begin:
!  Get the model dimension and common blocks.  
      include 'params.inc'
      include 'common.inc'
! Allocate array dimensions etc. 
      real t(NX,NY,0:NZ+1),to(NX,NY,0:NZ+1)
     .,    s(NX,NY,0:NZ+1),so(NX,NY,0:NZ+1)
      real r(NX,NY,0:NZ+1),p0(NX,NY),p(NX,NY,NZ+1)
     .,    u(NX,NY,NZ+1),v(NX,NY,NZ+1),w(NX,NY,0:NZ+1)
      real tt(NX,NY,NZ+1),st(NX,NY,NZ+1)
!*********************************************************************
! inserted by S.-K. Lee (May/19/2009)
      integer ntime
!*********************************************************************
!  Initialize

      call initlz(time0,t,to,s,so,r,p0,p,u,v,w,tt,st)

      iprt=max(1,nt/1000)  ! Print out values every iprt steps. 

!  Begin main (thermocline) integration loop
!  nt is number of timesteps, set in const.name
      do it=1,nt        
       time=time0+float(it)*dt            ! time counter
       if(ipr.ge.0 .and. mod(it,iprt).eq.0) then
        write(6,'(1x,i9,1x,f10.5,$)') it,time
       endif

!    Step t (and s) forward

       call pgstep(it,time,t,to,s,so,r,p0,p,u,v,w,tt,st)
       if(ipr.gt.0 .and. mod(it,iprt).eq.0)
     . write(6,'('' pgstep'',$)')
        if(ipr.ge.0 .and. mod(it,iprt).eq.0) then
         call diags(t,s,r,p,p0,u,v,w,diag1,diag2,diag3)
         write(6,'(3(1x,e13.6),$)') diag1,diag2,diag3
       endif
       if(ipr.ge.0 .and. mod(it,iprt).eq.0) write(6,'(1x)')
      enddo

!    End of main time integration loop

!    Write model state 

      call pguvw(time,t,s,r,p0,p,u,v,w)
!*********************************************************************
! inserted by S.-K. Lee (May/19/2009)
!     call pgout(nt,time,t,s,r,p0,p,u,v,w)
      ntime = int(time0)
      call pgout(ntime,time,t,s,r,p0,p,u,v,w)
!*********************************************************************

!  All done

      stop  'All done!'
      end
!
! ----------------------------------------------------------------
!
      subroutine pgstep(it,time,t,to,s,so,r,p0,p,u,v,w,tt,st)
!  This is the main timestepping routine. It calls a 
! number of other routines to evaluate terms in the 
! evolution equation, and then steps forward one timestep. 

!  Input arrays: t, to, s, so, p0
!  Arrays changed on output: t, to, s, so, r,  p, u, v, w, tt, st

! This routine calls routines pguvw, pgadv, ekman, eustpr, leapfr, 
!                             fluxbc, pgvdf, boundb, pghdf


!  Timestepping is determined by value of parameter irk, set in 
!  const.name, as follows:

! irk=0  Pseudo-Runge-Kutta (PRK) for advection, Euler Forward (EF)
!         for diffusion
! irk=1  PRK for both advection and diffusion.
! irk=2  2nd order Runge-Kutta (RK) for advection, EF for diffusion
! irk=3  RK for both advection and diffusion.

! The choice of which to use is left to the user. Considerations are: 
!  i) A forward step for diffusion is more stable than an RK step, and
!     therefore allows bigger timesteps if diffusion is a limitation. 
! ii) Choices 1 and 3 use the same stepping for diffusion and advection,
!     which may have some advantages for diagnostics. 
! iii)The RK method needs two evaluations of R.H.S. per timestep. The 
!     PRK method (sometimes called 'predictor-partial corrector')
!     needs only one, and therefore is almost twice as fast. 
!     Some accuracy may be lost. 

! The most economical timestepping scheme is therefore irk=0. 
! For more accurate stepping irk=1 or irk=2 may be an improvement. 
! irk=3 is generally hardly more accurate than irk=1 and needs 
! shorter timesteps than irk=2, so normally there is no need to use it,
! but is included for completeness and as an accuracy 'benchmark.'

! One strategy is to use irk=0 for the spin-up to equilibrium, and then
! a short period of irk=1 or irk=2 to evaluate diagnostics.

      include 'params.inc'
      include 'common.inc'
      real t(NX,NY,0:NZ+1),to(NX,NY,0:NZ+1)
     .,    s(NX,NY,0:NZ+1),so(NX,NY,0:NZ+1)
      real r(NX,NY,0:NZ+1),p0(NX,NY),p(NX,NY,NZ+1)
      real u(NX,NY,NZ+1),v(NX,NY,NZ+1),w(NX,NY,0:NZ+1)
      real tt(NX,NY,NZ+1),st(NX,NY,NZ+1)
 
! 
! First ever timestep (it.eq.1) is always RK (2nd order) or RK/EF
! 

! Predictor:
      if(it.eq.1 .or. irk.ge.2) then    ! Evaluate (RK only)
!   Get thermocline velocities from thermocline state
       call pguvw(time,t,s,r,p0,p,u,v,w)
       call zero(tt,NX,NY,NZ+1)
       call pgadv(t,u,v,w,tt)
       call ekman(time,t,u,v,w,gammats,ts,tt,1)
       if(isal.eq.1) then
        call zero(st,NX,NY,NZ+1)
        call pgadv(s,u,v,w,st)
        call ekman(time,s,u,v,w,gammass,ss,st,2)
       endif
       if(irk.eq.1. or. irk.eq.3) then   ! Diffusion (full RK only)
        call fluxbc(t,akvt,gammatb,tb0)
        call pgvdf(t,tb0,ts,akvt,gammats,gammat0,tt)
        call boundb(t)
        call pghdf(t,tt,akht,alht) 
        if(isal.eq.1) then
         call fluxbc(s,akvs,gammasb,sb0)
         call pgvdf(s,sb0,ss,akvs,gammass,gammas0,st)
         call boundb(s)
         call pghdf(s,st,akhs,alhs) 
        endif
       endif                             ! end full RK 
      endif                              ! end first RK evaluations
      
!%    if(irk.eq.0) call pguvw(time,t,s,r,p0,p,u,v,w)
      call eustpr(t,to,dt,tt)               ! T predictor step
      if(isal.eq.1) call eustpr(s,so,dt,st) ! S predictor step

! Euler Forward diffusion step:
      if(irk.eq.0. or. irk.eq.2) then  ! Euler Forward diffusion
       call zero(tt,NX,NY,NZ+1)
       call fluxbc(t,akvt,gammatb,tb0)
       call pgvdf(t,tb0,ts,akvt,gammats,gammat0,tt)
       call boundb(t)
       call pghdf(t,tt,akht,alht) 
       call eustpr(to,to,dt,tt)                
       if(isal.eq.1) then
        call zero(st,NX,NY,NZ+1)
        call fluxbc(s,akvs,gammasb,sb0)
        call pgvdf(s,sb0,ss,akvs,gammass,gammas0,st)
        call boundb(s)
        call pghdf(s,st,akhs,alhs) 
        call eustpr(so,so,dt,st)                
       endif
      endif                         ! end EF diffusion

! The first part of the timestep (predictor) is complete. 
! The new T and S are contained in to and so; t and s are unchanged.

!%     if(irk.ne.0)then 
!  Get new thermocline velocities from new thermocline state
        call pguvw(time,to,so,r,p0,p,u,v,w)
!%     endif

! Corrector:
      if(irk.eq.0 .or. irk.eq.2) then  ! Euler Forward diffusion step
       call eustpr(t,t,dt,tt)             
       call zero(tt,NX,NY,NZ+1)
       if(isal.eq.1) then
        call eustpr(s,s,dt,st)             
        call zero(st,NX,NY,NZ+1)
       endif
      endif                            ! end EF diffusion

      if(irk.eq.1 .or. irk.eq.3) then   ! RK/PRK diffusion (re-evaluate)
       call zero(tt,NX,NY,NZ+1)
       call fluxbc(to,akvt,gammatb,tb0)
       call pgvdf(to,tb0,ts,akvt,gammats,gammat0,tt)
       call boundb(to)
       call pghdf(to,tt,akht,alht) 
       if(isal.eq.1) then
        call zero(st,NX,NY,NZ+1)
        call fluxbc(so,akvs,gammasb,sb0)
        call pgvdf(so,sb0,ss,akvs,gammass,gammas0,st)
        call boundb(so)
        call pghdf(so,st,akhs,alhs) 
       endif
      endif                              ! end RK/PRK diffusion
 
! Now re-evaluate advection terms, for all cases
       call pgadv(to,u,v,w,tt)
       call ekman(time,to,u,v,w,gammats,ts,tt,1)
       call eustpr(t,t,dt,tt)   ! T advection corrector step 
       if(isal.eq.1) then
        call pgadv(so,u,v,w,st)
        call ekman(time,so,u,v,w,gammass,ss,st,2)
        call eustpr(s,s,dt,st)  ! S advection corrector step
       endif

! Predictor and corrector steps completed. Now average results:
      do k=1,NZ+1
       do j=2,NY-1
        do i=2,NX-1
         t(i,j,k)=0.5*(t(i,j,k)+to(i,j,k))
        enddo
       enddo
      enddo
      if(isal.eq.1) then
       do k=1,NZ+1
        do j=2,NY-1
         do i=2,NX-1
          s(i,j,k)=0.5*(s(i,j,k)+so(i,j,k))
         enddo
        enddo
       enddo
      endif

! Do a single convective adjustment if iconv=1
      if(iconv.eq.1) call convct(t,s)

! New T and S are now contained in t and s.
! For PRK, the corrector derivatives are stored in tt and st for use
!  at the next time-step. 

      return
      end
!
! ----------------------------------------------------------------------
!
      subroutine pgadv(r,u,v,w,xrt)
!   Temperature and salinity advection
! (r is either temperature or salinity, depending on call).
! Variables input: r,u,v,w,xrt
! Variables changed on output: xrt
! The calculated advected terms are added to the value of xrt on input.
! Uses common blocks.
! Advection is done in flux form with centered differences.
      include 'params.inc'
      include 'common.inc'
      real r(NX,NY,0:NZ+1),u(NX,NY,NZ+1),v(NX,NY,NZ+1),w(NX,NY,0:NZ+1)
      real xrt(NX,NY,NZ+1)
      real xur(NX-1,2:NY-1),xvr(2:NX-1,NY-1),xwr(0:NZ)
!  Horizontal advection
!   Use velocities in same form as in barotropic elliptic solver
!   Calculate fluxes first to avoid duplication
!      (Factor of 0.5 for r avg is picked up in r2dx, r2dy)
      do 100 k=1,NZ
       do 50 j=2,NY-1
        do 50 i=1,NX-1
         xur(i,j)=u(i,j,k)*(r(i+1,j,k)+r(i,j,k))*r2dx
50     continue
       do 60 j=1,NY-1
        do 60 i=2,NX-1
         xvr(i,j)=v(i,j,k)*(r(i,j+1,k)+r(i,j,k))*csv(j)*r2dy
60     continue
       do 80 j=2,NY-1
        do 80 i=2,NX-1
         xrt(i,j,k)=scu(j)*(xur(i-1,j)-xur(i,j)+xvr(i,j-1)-xvr(i,j))
     .               + xrt(i,j,k)
80     continue
100   continue
!  Vertical advection
!   Use fixed forcing velocity in w(i,j,NZ+1) rather than w(i,j,NZ)
!    for w_E. (They should be equal.)  Upwind scheme as in pguvw for
!    top advection.   Calculate fluxes first to avoid duplication
!      (Factor of 0.5 for r avg is picked up in r2dzr).
      do 200 j=2,NY-1
       do 200 i=2,NX-1
        xwr(0)=w(i,j,0)*(r(i,j,0)+r(i,j,0))
        do 120 k=1,NZ-1
         xwr(k)=w(i,j,k)*(r(i,j,k)+r(i,j,k+1))
120     continue
        if(w(i,j,NZ).le.0.) then
!        if(w(i,j,NZ+1).le.0.) then
         xwr(NZ)=w(i,j,NZ)*(r(i,j,NZ+1)+r(i,j,NZ+1))
!         xwr(NZ)=w(i,j,NZ+1)*(r(i,j,NZ+1)+r(i,j,NZ+1))
        else
         xwr(NZ)=w(i,j,NZ)*(r(i,j,NZ)+r(i,j,NZ))
!         xwr(NZ)=w(i,j,NZ+1)*(r(i,j,NZ)+r(i,j,NZ))
        endif
        do 150 k=1,NZ
         xrt(i,j,k)=xrt(i,j,k)+(xwr(k-1)-xwr(k))*r2dzr(k)
150     continue
200   continue
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine pgvdf(r,r0,rs,akv,gamma,gamma0,xrt)
!  Calculates vertical diffusion and surface (bulk) fluxes 
!  This can be a limiter on the maximum timestep:
!   stable for Euler forward if 2*akv*dt/dz**2 < 1

!   Input variables:  r, xrt, akv
!   Variables changed on output: xrt 
!   Uses common blocks.

      include 'params.inc'
      include 'common.inc'
      real r(NX,NY,0:NZ+1),xrt(NX,NY,NZ+1)
      real rs(NX,NY)
!  Ekman layer and top interior point
      do j=2,NY-1
       do i=2,NX-1
        xrt(i,j,NZ+1)=-gamma0*rde*akv*(r(i,j,NZ+1)-r(i,j,NZ))*rdzw(NZ)
!*********************************************************************
! inserted by S.-K. Lee (May/19/2009)
!    .                -gamma*(r(i,j,NZ+1)-rs(i,j))*rde
     .                -gamma*(r0         -rs(i,j))*rde
!*********************************************************************
     .                +xrt(i,j,NZ+1)
        xrt(i,j,NZ)=akv*(gamma0*(r(i,j,NZ+1)-r(i,j,NZ))*rdzw(NZ)
     .                   -(r(i,j,NZ)-r(i,j,NZ-1))*rdzw(NZ-1))*rdzr(NZ)
     .                 +xrt(i,j,NZ)
       enddo
      enddo
!  Interior
!   Top and bottom differencing agrees with the formulation of the flux
!    conditions in ekman and fluxbc (but are only first-order accurate)
      do k=1,NZ-1
       do j=2,NY-1
        do i=2,NX-1
         xrt(i,j,k)=akv*((r(i,j,k+1)-r(i,j,k))*rdzw(k)
     .                  -(r(i,j,k)-r(i,j,k-1))*rdzw(k-1))*rdzr(k)
     .                 +xrt(i,j,k)
        enddo
       enddo
      enddo
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine pghdf(r,xrt,akh,alh)
! This routine evaluates both the bi-harmonic and laplacian diffusion
!  Bi-harmonic horizontal diffusion                    
!   is implemented as two Laplacian diffusion steps with an intermediate
!    calculation of boundary values from the no-flux condition

!   Input variables:  r, xrt, akh, alh
!   Variables changed on output: xrt 
!   Uses the work array xr0 (locally dimensioned and allocated). 

      include 'params.inc'
      include 'common.inc'
      real r(NX,NY,0:NZ+1),xrt(NX,NY,NZ+1),xr0(NX,NY)
!  One loop over depth, to reduce dimension of temporary array xr0
      do 100 k=1,NZ
!  First Laplacian step, with akh and both factors of rdx**2
       do j=2,NY-1
        do i=2,NX-1
!  First term is Laplacian term, second is first biharmonic step
         xr0(i,j)=akh*r(i,j,k)
     . -alh*scu(j)*(scu(j)*rdx**2*(r(i+1,j,k)-r(i,j,k)
     .                            -r(i,j,k)+r(i-1,j,k))
     .              +rdy**2*((r(i,j+1,k)-r(i,j,k))*csv(j)
     .                      -(r(i,j,k)-r(i,j-1,k))*csv(j-1)))
         enddo
        enddo

!  Set boundary values using no-flux condition
       do i=2,NX-1
        xr0(i,1)=xr0(i,2)
        xr0(i,NY)=xr0(i,NY-1)
       enddo
       do j=2,NY-1
        xr0(1,j)=xr0(2,j)
        xr0(NX,j)=xr0(NX-1,j)
       enddo
!  Second Laplacian step
       do j=2,NY-1
        do  i=2,NX-1
         xrt(i,j,k)=xrt(i,j,k)
     .   +scu(j)*(scu(j)*rdx**2*(xr0(i+1,j)-xr0(i,j)
     .                          -xr0(i,j)+xr0(i-1,j))
     .            +rdy**2*((xr0(i,j+1)-xr0(i,j))*csv(j)
     .                    -(xr0(i,j)-xr0(i,j-1))*csv(j-1)))
        enddo
       enddo
 100  continue
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine convct(t,s)
!  Convectively adjust to eliminate static instability
! This is a relatively simple convective adjustment scheme. 
! Its main current restriction is the assumption of a linear
! equation of state. 
! For each column (i.e. i,j location) the algorithm
! scans downward, testing for static instability. 
! Where instability is found, the profiles are adjusted to
! neutral stability (i.e. uniform profiles) conserving the 
! L_1 norm (the linear sum) of the variable. 

! Variables input: t, s
! Variables changed on output: t, s
! Also uses common block iconst, rconst, paramet. 

      include 'params.inc'
      include 'common.inc'
      real t(NX,NY,0:NZ+1),s(NX,NY,0:NZ+1)
      real t1(0:NZ+1),s1(0:NZ+1)

      do 100 j=2,NY-1
       do 100 i=2,NX-1
        do 2 k=1,NZ+1
         t1(k)=t(i,j,k)
2       continue
        if(isal.eq.1) then
         do 4 k=1,NZ+1
          s1(k)=s(i,j,k)
4        continue
        endif
!  Stratification is neutral for k: kb .le.  k .le. kt,
!   stable/neutral above kt, stable from kt to kt+1, unknown below kb
!  Step down until kb=1
        kt=NZ+1
        kb=kt
10      continue
!  Done if kb=1
         if(kb.le.1) goto 90
!  Stable at kt+1/kt
!  Check stability below kb
!   If stable at kb/kb-1, go and check the next kb
         if( dents1(t1(kb),s1(kb),isal)
     .    .lt. dents1(t1(kb-1),s1(kb-1),isal) ) goto 80
!   If unstable, mix down one; skip if neutral
          if( dents1(t1(kb),s1(kb),isal)
     .     .gt. dents1(t1(kb-1),s1(kb-1),isal) ) then
!     zt-zb is the vertical extent of the mixed fluid
!     zb-zb1 is the vertical extent of the fluid to be mixed
           if(kt.lt.NZ) then
            zt=0.5*(z(kt)+z(kt+1))
           else if(kt.eq.NZ) then
            zt=z(NZ)
!            zt=z(NZ+1)
           else
            zt=z(NZ+1)
!            zt=z(NZ+1)+de
           endif
           if(kb.eq.NZ+1) then
            zb=z(NZ)
!            zb=z(NZ+1)
           else
            zb=0.5*(z(kb-1)+z(kb))
           endif
           if(kb.ge.3) then
            zb1=0.5*(z(kb-2)+z(kb-1))
           else
            zb1=z(0)
           endif
           t1(kb-1)=((zt-zb)*t1(kb)+(zb-zb1)*t1(kb-1))/(zt-zb1)
           do 20 k=kb,kt
            t1(k)=t1(kb-1)
20         continue
           if(isal.eq.1) then
            s1(kb-1)=((zt-zb)*s1(kb)+(zb-zb1)*s1(kb-1))/(zt-zb1)
!            s1(kb-1)=(rmix*s1(kb)+s1(kb-1))/(rmix+1.)
            do 22 k=kb,kt
             s1(k)=s1(kb-1)
22          continue
           endif
          endif
          kb=kb-1
50        continue
!   Check stability above kt
!    If at the top, go and check the next kb
           if(kt.eq.NZ+1) goto 10
!    If stable at kt+1/kt, go and check the next kb
           if( dents1(t1(kt+1),s1(kt+1),isal)
     .      .lt. dents1(t1(kt),s1(kt),isal) ) goto 10
!   If unstable, mix up one; skip if neutral
           if( dents1(t1(kt+1),s1(kt+1),isal)
     .      .gt. dents1(t1(kt),s1(kt),isal) ) then
!     zt1-zt is the vertical extent of the fluid to be mixed
!     zt-zb is the vertical extent of the mixed fluid
            if(kt.lt.NZ-1) then
             zt1=0.5*(z(kt+1)+z(kt+2))
             zt=0.5*(z(kt)+z(kt+1))
            else if(kt.eq.NZ-1) then
             zt1=z(NZ)
!             zt1=z(NZ+1)
             zt=0.5*(z(NZ-1)+z(NZ))
            else
             zt1=z(NZ+1)
             zt=z(NZ)
!             zt1=z(NZ+1)+de
!             zt=z(NZ+1)
            endif
            if(kb.ge.2) then
             zb=0.5*(z(kb-1)+z(kb))
            else
              zb=z(0)
            endif
            t1(kt+1)=((zt-zb)*t1(kt)+(zt1-zt)*t1(kt+1))/(zt1-zb)
            do 60 k=kb,kt
             t1(k)=t1(kt+1)
60          continue
            if(isal.eq.1) then
             s1(kt+1)=((zt-zb)*s1(kt)+(zt1-zt)*s1(kt+1))/(zt1-zb)
!             s1(kt+1)=(rmix*s1(kt)+rmix1*s1(kt+1))/(rmix+rmix1)
             do 62 k=kb,kt
              s1(k)=s1(kt+1)
62           continue
            endif
           endif
           kt=kt+1
!   Go and check the next one up
          goto 50
80       continue
!  Stable at kb/kb-1 and kt+1/kt
         kb=kb-1
         kt=kb
        if(kb.gt.1) goto 10
!  Now all done
90      continue
        do 92 k=1,NZ+1
         t(i,j,k)=t1(k)
92      continue
        if(isal.eq.1) then
         do 94 k=1,NZ+1
          s(i,j,k)=s1(k)
94       continue
        endif
100   continue
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine dents(t,s,r)
! Evaluate linear equation of state r(t,s) everywhere, including
! side boundaries.
! Input arrays: s, t
! Variables changed on output: r 
      include 'params.inc'
      include 'common.inc'

      real t(NX,NY,0:NZ+1),s(NX,NY,0:NZ+1),r(NX,NY,0:NZ+1)
      if(isal.eq.1) then
       do  k=0,NZ+1
        do  j=1,NY
         do  i=1,NX
          r(i,j,k)=s(i,j,k)-t(i,j,k)
         enddo
        enddo
       enddo
      else
       do k=0,NZ+1
        do j=1,NY
         do i=1,NX
          r(i,j,k)=-t(i,j,k)
         enddo
        enddo
       enddo
      endif
      return
      end
!
! ---------------------------------------------------------------
!
      function dents2(t1,s1,isal)
!  Evaluate linear equation of state r(t,s) at one point
       dents2=s1*isal-t1
      return
      end
!
! ---------------------------------------------------------------
!
      function dents1(t1,s1,isal)
!  Evaluate linear equation of state r(t,s) at one point
      if(isal.eq.1) then
       dents1=s1-t1
      else
       dents1=-t1
      endif
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine ekman(time,r,u,v,w,gamma,rs,xrt,iflag)
!  Calculate Ekman layer t, s time-derivatives
!   iflag=1 => t, =2 => s (i.e. does temperature or salinity 
!   depending on whether iflag = 1 or iflag=2)
!   The Ekman layer velocities u,v(NZ+1) are set in pguvw
! Input variables: r,u,v,w,gamma,rs,xrt,iflag
! Variables changed on output:  xrt 
      include 'params.inc'
      include 'common.inc'
      real r(NX,NY,0:NZ+1),u(NX,NY,NZ+1),v(NX,NY,NZ+1),w(NX,NY,0:NZ+1)
      real xrt(NX,NY,NZ+1),rs(NX,NY)

      do 10 j=2,NY-1
       do 10 i=2,NX-1

!  Advection--upwind scheme
        if(u(i-1,j,NZ+1).gt.0.) then
         fx1=u(i-1,j,NZ+1)*r(i-1,j,NZ+1)
        else
         fx1=u(i-1,j,NZ+1)*r(i,j,NZ+1)
        endif
        if(u(i,j,NZ+1).gt.0.) then
         fx2=u(i,j,NZ+1)*r(i,j,NZ+1)
        else
         fx2=u(i,j,NZ+1)*r(i+1,j,NZ+1)
        endif
        if(v(i,j-1,NZ+1).gt.0.) then
         fy1=v(i,j-1,NZ+1)*r(i,j-1,NZ+1)
        else
         fy1=v(i,j-1,NZ+1)*r(i,j,NZ+1)
        endif
        if(v(i,j,NZ+1).gt.0.) then
         fy2=v(i,j,NZ+1)*r(i,j,NZ+1)
        else
         fy2=v(i,j,NZ+1)*r(i,j+1,NZ+1)
        endif
        xrt(i,j,NZ+1)=xrt(i,j,NZ+1)+scu(j)*( (fx1-fx2)*rdx
     .        +(fy1*csv(j-1)-fy2*csv(j))*rdy )
10    continue
!        xrt(i,j,NZ+1)=(fx1-fx2)*rdx+(fy1-fy2)*rdy

!  Vertical advection/surface flux
!      (Vertical diffusion is calculated in pgvdf)
      do 100 j=2,NY-1
       do 100 i=2,NX-1
!   Upwind vertical advection
        if(w(i,j,NZ).le.0.) then
         xrt(i,j,NZ+1)=xrt(i,j,NZ+1)+w(i,j,NZ)*r(i,j,NZ+1)*rde
        else
         xrt(i,j,NZ+1)=xrt(i,j,NZ+1)+w(i,j,NZ)*r(i,j,NZ)*rde
        endif
100   continue
!  Freshwater flux
      if(iflag.eq.1) then
!   Temperature
       do 200 j=2,NY-1
        do 200 i=2,NX-1
         if(w0(i,j).le.0.) then
          xrt(i,j,NZ+1)=xrt(i,j,NZ+1)-w0(i,j)*ts(i,j)*rde
         else
          xrt(i,j,NZ+1)=xrt(i,j,NZ+1)-w0(i,j)*r(i,j,NZ+1)*rde
         endif
200    continue
      else if(iflag.eq.2) then
!   Salinity
       do 210 j=2,NY-1
        do 210 i=2,NX-1
         xrt(i,j,NZ+1)=xrt(i,j,NZ+1)+w0(i,j)*scap0*rde
210    continue
      endif
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine eustpr(ro,r,dt,rt)
! This routine does a simple Euler timestep
! Input variables: ro, dt, rt
! Variable changed on output: r
      include 'params.inc'
      real ro(NX,NY,0:NZ+1),r(NX,NY,0:NZ+1),rt(NX,NY,NZ+1)
!  Euler time step
      do  k=1,NZ+1
       do  j=2,NY-1
        do  i=2,NX-1
         r(i,j,k)=ro(i,j,k)+dt*rt(i,j,k)
        enddo
       enddo
      enddo
      return
      end
!
! ----------------------------------------------------------------
!
      subroutine fluxbc(r,akv,gammab,rb0)
!  Calculate bottom (z=0) temperature and salinity
!   for advection/diffusion using flux condition
! Input variables: r, akv, gammab, rb0
! Variables changed on output: r. 
      include 'params.inc'
      include 'common.inc'
      real r(NX,NY,0:NZ+1)
!   Diffusive flux condition:
      do j=2,NY-1
       do i=2,NX-1
         r(i,j,0)=(2.*akv*r(i,j,1)+gammab*(z(1)-z(0))*rb0)
     .           *(1./(2.*akv+gammab*(z(1)-z(0))))
       enddo
      enddo
      return
      end

! ----------------------------------------------------------------
!
      subroutine leap(ro,r,dt2,rt)
! This routine does a forward timestep
! as part of a leapfrog scheme.  Tests of a leapfog scheme 
! showed that longer timesteps were needed, so the scheme
! was dropped as an option.
! Uses 'Robert filter' if rob is non-zero. 
! Input variables: ro, dt2, rt
! Variable changed on output: r, ro
      include 'params.inc'
      real ro(NX,NY,0:NZ+1),r(NX,NY,0:NZ+1),rt(NX,NY,NZ+1)
      rob=0.02   ! parameter for Robert filter to prevent splitting
!  Time step:
      do  k=1,NZ+1
       do  j=2,NY-1
        do  i=2,NX-1
         tmpr=r(i,j,k)
         r(i,j,k)=ro(i,j,k)+dt2*rt(i,j,k)
         ro(i,j,k)=tmpr + rob*(ro(i,j,k) + r(i,j,k) - 2*tmpr)
        enddo
       enddo
      enddo
      return
      end
!
! ----------------------------------------------------------------
! 
	subroutine zero(a,n1,n2,n3)
        real a(n1,n2,n3)
        do k=1,n3
         do j=1,n2
           do i=1,n1
             a(i,j,k)=0.
            enddo
          enddo
         enddo
         return
         end
!
! ----------------------------------------------------------------
!
